package Boulder::Genbank;
use Boulder::Stream;

require Exporter;
@ISA = qw(Exporter Boulder::Stream);
@EXPORT = ();
@EXPORT_OK = ();

use Carp;

# Genbank entry parsing constants
# (may need to adjust!)
$KEYCOL=0;
$VALUECOL=12;
$FEATURECOL=5;
$FEATUREVALCOL=21;


# new() takes named parameters:
# -accessor=> Reference to an object class that will return a series of
#           Genbank records.  Predefined objects include 'Yank', 'NCBI', and 'Entrez'.
#           (defaults to 'Yank').
# -param=>  Parameters to pass to the subroutine.  Can be a list of accession numbers
#           or an entrez query.
# -out=>    Output filehandle.  Defaults to STDOUT.
#
# If you don't use named parameters, then will assume method 'yank' on
# a list of accession numbers.
# e.g.
#        $gb = new Boulder::Genbank(-accessor=>'Yank',-param=>[qw/M57939 M28274 L36028/]);
sub new {
    my($package,@parameters) = @_;
    # superclass constructor
    my($self) = new Boulder::Stream;
    
    # figure out whether parameters are named.  Look for
    # an initial '-'
    if ($parameters[0]=~/^-/) {
	my(%parameters) = @parameters;
	$self->{'accessor'}=$parameters{'-accessor'} || 'Yank';
	$self->{'param'}=$parameters{'-param'};
	$self->{'OUT'}=$parameters{'-out'} || 'main::STDOUT';
    } else {
	$self->{'accessor'}='Yank';
	$self->{'param'}=[@parameters];
    }
    
    croak "Require parameters" unless defined($self->{'param'});
    my(@param) = ref($self->{'param'}) eq 'ARRAY' ? 
	@{$self->{'param'}} : $self->{'param'};
    $self->{'accessor'} = new {$self->{'accessor'}}(@param);
    
    return bless $self,$package;
}

sub read_record {
    my($self,@tags) = @_;
    my($s);

    if (wantarray) {
	my(@result);
	while (!$self->{'done'}) {
	    $s = $self->read_one_record(@tags);
	    next unless $s;
	    next if $query && !(&$query);
	    push(@result,$s);
	}
	return @result;
    } 

    # we get here if in a scalar context
    while (!$self->{'done'}) {
	$s = $self->read_one_record(@tags);
	next unless $s;
	return $s unless $query;
	return $s if &$query;
    }
    return undef;
}

sub read_one_record {
    my($self,@tags) = @_;
    my(%ok);
    grep($ok{$_}++,@tags);
    
    my($accessor) = $self->{'accessor'};
    my($record) = $accessor->fetch_next();
    unless ($record) {
	$self->{'done'}++;
	return undef;
    }
    
    my($s,@lines,$line,$accumulated,$key,$keyword,$value,$feature,@features);

    $s = new Stone;
    @lines = split("\n",$record);

    foreach $line (@lines) {

	# special case for the sequence itself
	if ($line=~/^ORIGIN/) {
	    $self->_addToStone($key,$accumulated,$s,\%ok) if $key;
	    last;
	}

	# special case for the features table
	if ($line=~/^FEATURES/..$line=~/^BASE COUNT/) {
	    undef $keyword;

	    if ($line=~/^FEATURES/) {
		undef @features;
		next;
	    }

	    if ($line=~/BASE COUNT/) {
		push(@features,$feature) if $feature;
		$self->_addFeaturesToStone(\@features,trim($'),$s,\%ok);
		next;
	    }

	    my($featurelabel) = trim(substr($line,$FEATURECOL,$FEATUREVALCOL-$FEATURECOL));
	    my($featurevalue) = trim(substr($line,$FEATUREVALCOL));
	    if ($featurelabel) {
		push(@features,$feature) if $feature;
		$feature = {'label'=>$featurelabel,'value'=>$featurevalue};
	    } else {
		$feature->{'value'} .= " $featurevalue";
	    }

	    next;
	}

	$keyword = trim(substr($line,0,$VALUECOL-1));
	$value = trim(substr($line,$VALUECOL));

	if ($keyword && $key) {
	    $self->_addToStone($key,$accumulated,$s,\%ok);
	    $accumulated = $value;
	    next;
	}
	$accumulated .= " $value";
    } continue {
	$key = $keyword if $keyword;
    }
    
    ($sequence)=$record=~/\nORIGIN.*\n([\s\S]+)/;
    $sequence=~tr/gatcnGATCN//cd;
    $self->_addToStone('SEQUENCE',$sequence,$s,\%ok);
    return $s;
}


sub trim {
    my($v) = @_;
    $v=~s/^\s+//;
    $v=~s/\s+$//;
    return $v;
}

sub _addToStone {
    my($self,$label,$value,$stone,$ok) = @_;
    return unless !%{$ok} || $ok->{$label};
    $stone->insert($label,$value);
}

sub _addFeaturesToStone {
    my($self,$features,$basecount,$stone,$ok) = @_;

    # first add the basecount
    if (!%{$ok} || $ok->{'BASECOUNT'}) {
	my(%counts) = $basecount=~/(\d+)\s+([gatcGATC])/g;
	%counts = reverse %counts;
	$stone->insert('BASECOUNT',new Stone(%counts));
    }
    
    if (!%{$ok} || $ok->{'FEATURES'}) {
	# now add the features
	my($f) = new Stone;
	foreach (@$features) {
	    my($q) = $_->{'value'};
	    my($label) = $_->{'label'};
	    my($position) = $q=~m!^([^/]+)!;
	    my(%qualifiers) = $q=~m!/(\w+)=([^/]+)!g;
	    # strip off the (unnecessary) quotation marks
	    foreach (keys %qualifiers) {
		$qualifiers{$_}=~s/^\s*"//;
	    $qualifiers{$_}=~s/"\s*$//;	# 
	    }
	    $f->insert($label=>new Stone('position'=>$position,%qualifiers));
	}
	$stone->insert('FEATURES',$f);
    }
}

# ----------------------------------------------------------------------------------------
# -------------------------- DEFINITION OF ACCESSOR OBJECTS ------------------------------
package GenbankAccessor;
use Carp;

sub new {
    my($class,@parameters) = @_;
    croak "GenbankAccessor::new:  Abstract class\n";
}

sub fetch_next {
    my($self) = @_;
    croak "GenbankAccessor::fetch_next: Abstract class\n";
}

sub DESTROY {
}

package Yank;
use Carp;

@ISA=qw(GenbankAccessor);
$YANK='/usr/local/bin/yank';

sub new {
    my($package,@accession) = @_;
    croak "Yank::new(): need at least one Genbank acccession number" unless @accession;
    croak "Yank::new(): yank executable not found" unless -x $YANK;
    my($tmpfile) = "/usr/tmp/yank$$";
    open (TMP,">$tmpfile") || croak "Yank::new(): couldn't open tmpfile $tmpfile for write: $!";
    print TMP join("\n",@accession),"\n";
    close TMP;
    open(YANK,"$YANK < $tmpfile |") || croak "Yank::new(): couldn't open pipe from yank: $!";
    return bless {'tmpfile'=>$tmpfile,'fh'=>YANK},$package;
}

sub fetch_next {
    my($self) = @_;
    return undef unless $self->{'fh'};
    local($/) = "//\n";
    my($line);
    my($fh) = $self->{'fh'};
    chomp($line = <$fh>);
    return $line;
}

sub DESTROY {
    my($self) = shift;
    close $self->{'fh'} if $self->{'fh'};
    unlink $self->{'tmpfile'} if $self->{'tmpfile'}
}

package File;
use Carp;
@ISA=qw(GenbankAccessor);
$DEFAULT_PATH = '/net/draino/local2/genbank/gbpri.seq';

sub new {
    my($package,$path) = @_;
    $path = $DEFAULT_PATH unless $path;
    open (GB,$path) or croak "File::new(): couldn't open $path: $!";
    # read the junk at the beginning
    my $found;
    while (<GB>) {
	$found++,last if /^\s*\d+ loci/;
    }
    croak "File::new(): $path doesn't look like a GenBank flat file"
	unless $found;
    $_ = <GB>;
    return bless {'fh'=>GB},$package;
}

sub fetch_next {
    my $self = shift;
    return undef unless $self->{'fh'};
    local($/)="//\n";
    my($line);
    my($fh) = $self->{'fh'};
    chomp($line = <$fh>);
    return $line;
}

package Entrez;
use Carp;
use LWP::Simple;

@ISA=qw(GenbankAccessor);
$GENBANK_URL="http://www3.ncbi.nlm.nih.gov:80/htbin-post/Entrez/query?form=6&db=n&Dopt=g&uid=";

sub new {
    my($package,@accession) = @_;
    croak "Entrez::new(): need at least one Genbank acccession number" unless @accession;
    return bless {'accession'=>[@accession]},$package;
}

sub fetch_next {
    my($self) = @_;

    # if any additional records are left, then return them
    return shift(@{$self->{'records'}})
	if @{$self->{'records'}};

    my($accession) = shift @{$self->{'accession'}};
    return undef unless $accession;
  LWP::Debug::level('+');
    my($content) = get("$GENBANK_URL$accession");

    # trim off the junk at the top
    $content=~s/\A[\s\S]\nDEFINITION/DEFINITION/m;
    # trim off the cruft at the bottom
    $content=~s![\n\s]+</pre>.*\Z!!;

    # remove HTML anchors
    $content=~s!<a.*>([^<>]+)</a>!$1!g;
    $content=~s/&lt;/</g;
    $content=~s/&gt;/</g;
    $self->{'records'}=[split("//\n",$content)];
    return shift @{$self->{'records'}};
}


1;
